

## logging with dates
# JL Documentation page rendering may not compatible with LogginExtras, so use timestamp instead.
timestamp() = Dates.format(now(), "yyyy-mm-dd HH:MM:SS ")

## UUID generation
const UUID4 = uuid4(UUIDs.MersenneTwister(39871))

## simple functions
do_nothing() = nothing
do_nothing(x) = nothing
do_nothing(x, y) = nothing

isok(x::Nothing) = true
isok(x::Bool) = x
isok(x::AbstractString) = !isempty(x) && !occursin(r"^n(o|ull)?$|^f(alse)?$|^0$"i, x)
isok(x) = true  # default is true

## parse default inputs/outputs
function throw_invalid_xxputs(any)
	throw(ErrorException("TypeError: Elements of inputs and outputs can only be one of `name::Union{String,Symbol}`, `name => default_value`, `name => data_type`, `name => default_value => data_type`, `name => data_type => default_value`. Invalid value: $any"))
end

"""
    parse_default(v)

Parsing `inputs` and `outputs` when creating `Program` objects.

Return `xxputs::Vector{String}, xxput_types::Vector{Type}, default_xxputs::Vector`.

## Valid `v` element types

- `name::String`: no default value.

- `name::String => value`: set default value, except `value` is `nothing` (default value not set).

- `name::String => value_type::Type`: no default value, but value type.

- `name::String => value => value_type::Type`: set default value and value type.

- `name::String => value_type::Type => value`: set default value and value type.
"""
function parse_default(v::Vector{String})
    xxput_types = Type[Any for i = 1:length(v)]
    default_xxputs = Nothing[nothing for i = 1:length(v)]
    return v, xxput_types, default_xxputs
end

function parse_default(v::Vector)
    n = length(v)
    xxputs = Vector{String}(undef, n)
    xxput_types = Vector{Type}(undef, n)
    default_xxputs = Vector{Any}(undef, n)
    for (i, ele) in enumerate(v)
        name, type, default = parse_default_element(ele)
        xxputs[i] = name
        xxput_types[i] = type
        default_xxputs[i] = default
    end
    return xxputs, xxput_types, default_xxputs
end

parse_default(s::String) = parse_default([s])
parse_default(s::AbstractString) = parse_default([str(s)])
parse_default(p::Pair) = parse_default([p])
parse_default(d::Dict) = parse_default([p for p in d])
parse_default(s::Symbol) = parse_default([str(s)])
parse_default(any) = throw_invalid_xxputs(any)

function parse_default_element(ele::String)
    return ele, Any, nothing
end
function parse_default_element(ele::AbstractString)
    return str(ele), Any, nothing
end
function parse_default_element(ele::Symbol)
    return str(ele), Any, nothing
end
function parse_default_element(ele::Pair)
	if ele.first isa AbstractString || ele.first isa Symbol
	    name = str(ele.first)
		data_type, value = parse_arg_info(ele.second)
		name, data_type, value
	else
		throw_invalid_xxputs(ele)
	end
end
parse_default_element(any) = throw_invalid_xxputs(any)

function parse_arg_info(data_type::Type)
	data_type, nothing
end
function parse_arg_info(p::Pair)
	parse_arg_info_pair(p.first, p.second)
end
function parse_arg_info(value)
	Any, value
end
function parse_arg_info_pair(a::Type, b::Type)
	if a isa b
		b, a
	elseif b isa a
		a, b
	else
		throw(ErrorException("TypeError: $(a) and $(b) are not inclusive."))
	end
end
parse_arg_info_pair(a, b::Type) = b, convert_data_type(a, b)
parse_arg_info_pair(a::Type, b) = a, convert_data_type(b, a)


function convert_data_type(value, data_type::Type)
	if isa(value, data_type)
		return value
	else
		try
			convert(data_type, value)
		catch
			throw(ErrorException("TypeError: cannot convert $value to $data_type."))
		end
	end
end

## convert inputs/outputs to Dict{String} in `run(p, inputs, outputs)`
function to_xxput_dict(p::Pair{String, V}) where V
	Dict(p.first => p.second)
end
function to_xxput_dict(p::Pair)
	Dict(str(p.first) => p.second)
end
function to_xxput_dict(v::Vector{V}) where V <: Pair
	res = Dict{String,Any}()
	for p in v
		res[str(p.first)] = p.second
	end
	res
end
function to_xxput_dict(d::Dict)
	res = Dict{String,Any}()
	for p in v
		res[str(p.first)] = p.second
	end
	res
end
to_xxput_dict(d::Dict{String}) = d
to_xxput_dict(any) = throw(ErrorException("TypeError: cannot run Program: cannot convert to Dict{String}: inputs, outputs, or returned value of infer_outputs(inputs)"))


## String/Cmd conversion
"""
    to_str(x) -> String
    str(x) -> String

Convert `x` to `String`.

- `x::Cmd`: remove backticks (return `string(x)[2:end-1]`).

- `x::Nothing`: return `""`.

- `x::Vector`: join elements with `"_"` as delim.

- `x::Any`: return `string(x)`.
"""
str(x::Cmd) = string(x)[2:end-1]
str(::Nothing) = ""
str(x::Vector) = join(str.(x), "_")
str(x::Regex) = x.pattern
str(any) = string(any)
to_str = str

"""
    to_cmd(x) -> Cmd

Convert `x` to Cmd.

Exception: when `x::Nothing`, return `nothing::Nothing`.
"""
to_cmd(c::Cmd) = c
to_cmd(s::AbstractString) = Cmd([str(s)])
to_cmd(v::Vector{String}) = Cmd(v)
to_cmd(v::Vector{T}) where T = Cmd(str.(v))
to_cmd(::Nothing) = nothing
to_cmd(i::Number) = Cmd([str(i)])
to_cmd(any) = Cmd([str(any)])

ValidInputTypes = Union{Cmd, AbstractString, Vector, Number}

"""
    split(c::Cmd) = c.exec :: Vector{String}

Return splitted arguments of `Cmd`.
"""
function Base.split(c::Cmd)
    c.exec
end

## path functions
"""
    replaceext(path, replacement::AbstractString)

If the last component of a path contains a dot, leave everything before the dot as usual, and everything after the dot is replaced by `replacement`. Otherwise, `replacement` will be appended to `path`.

If `replacement` is empty, the last dot will be removed.
"""
function replaceext(path::String, replacement::AbstractString)
    a, b = splitext(path)
    if isempty(replacement)
        a
    elseif replacement[1] == '.'
        a * replacement
    else
        a * "." * replacement
    end
end
replaceext(path, replacement::AbstractString) = replaceext(to_str(path), replacement::String)

"""
    removeext(path)

If the last component of a path contains a dot, leave everything before the dot as usual, and everything including and after the dot is discarded.
"""
removeext(path::String) = splitext(path)[1]
removeext(path) = removeext(to_str(path))
